import { HttpException, HttpService, Injectable, HttpStatus, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Types, Model } from 'mongoose';
import { Observable } from 'rxjs';
import { AxiosResponse } from 'axios';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import * as ejs from 'ejs';
import { structureDto } from '../dto/structure.dto';
import { UsersService } from '../../users/users.service';
import { User } from '../../users/schemas/user.schema';
import { MailerService } from '../../mailer/mailer.service';
import { Cron, CronExpression } from '@nestjs/schedule';
import { DateTime } from 'luxon';

@Injectable()
export class StructuresService {
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    private readonly mailerService: MailerService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  public async create(idUser: string, structureDto: structureDto): Promise<Structure> {
    const user = await this.userService.findOne(idUser);
    if (!user) {
      throw new HttpException('Invalid profile', HttpStatus.NOT_FOUND);
    }
    const createdStructure = new this.structureModel(structureDto);
    createdStructure._id = Types.ObjectId();
    createdStructure.save();
    user.structuresLink.push(createdStructure._id);
    user.save();

    return createdStructure;
  }

  public async search(searchString: string, filters?: Array<any>): Promise<Structure[]> {
    if (searchString && filters) {
      return this.structureModel
        .find({
          $and: [...this.parseFilter(filters), { $text: { $search: searchString }, deletedAt: { $exists: false } }],
        })
        .exec();
    } else if (filters) {
      return this.structureModel
        .find({ $and: [{ $or: this.parseFilter(filters), deletedAt: { $exists: false } }] })
        .exec();
    } else {
      return this.structureModel
        .find({ $and: [{ $or: [{ $text: { $search: searchString }, deletedAt: { $exists: false } }] }] })
        .exec();
    }
  }

  /**
   * Parse filter value from string to boolean
   * @param filters
   */
  private parseFilter(filters: Array<any>): Array<any> {
    return filters.map((filter) => {
      const key = Object.keys(filter)[0];
      if (filter[key] === 'True') {
        return { [key]: true };
      } else {
        return filter;
      }
    });
  }

  public async findAll(): Promise<StructureDocument[]> {
    const structures = await this.structureModel.find({ deletedAt: { $exists: false } }).exec();
    // Update structures coord and address before sending them
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        // If structre has no address, add it
        if (!structure.address) {
          return this.getStructurePosition(structure).then((postition) => {
            this.structureModel
              .findByIdAndUpdate(Types.ObjectId(structure._id), { address: postition.address, coord: postition.coord })
              .exec();
          });
        }
        if (structure.coord.length <= 0) {
          return new Promise((resolve) => {
            this.getStructurePosition(structure).then((postition: StructureDocument) => {
              this.structureModel
                .findByIdAndUpdate(Types.ObjectId(postition._id), { coord: postition.coord })
                .exec()
                .then(() => {
                  resolve('');
                });
            });
          });
        }
      })
    );
    return this.structureModel.find({ deletedAt: { $exists: false } }).exec();
  }

  public async update(idStructure: string, structure: structureDto): Promise<Structure> {
    const result = await this.structureModel.findByIdAndUpdate(Types.ObjectId(idStructure), structure).exec();
    if (!result) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    return this.findOne(idStructure);
  }

  public async findOne(idParam: string): Promise<Structure> {
    return await this.structureModel.findById(Types.ObjectId(idParam)).exec();
  }
  /**
   * Get structures positions and add marker corresponding to those positons on the map
   */
  private getStructurePosition(structure: Structure): Promise<Structure> {
    return new Promise((resolve) => {
      this.getCoord(structure.address.numero, structure.address.street, structure.address.commune).subscribe(
        (res) => {
          const address = res.data.features[0];
          structure.coord = address.geometry.coordinates;
          resolve(structure);
        },
        (err) => {
          Logger.error(`Request error: ${err.config.url}`, 'StructureService');
          Logger.error(err);
        }
      );
    });
  }

  public async isClaimed(structureId: string, user: User): Promise<boolean> {
    const isStructureClaimed = await this.userService.isStructureClaimed(structureId.toString());
    const isUserAlreadyClaimed = await this.userService.isUserAlreadyClaimedStructure(structureId, user.email);
    if (isStructureClaimed || isUserAlreadyClaimed) {
      return true;
    }
    return false;
  }

  /**
   * Search structure address based on data search WS
   */
  public async searchAddress(data: { searchQuery: string }): Promise<AxiosResponse<any>> {
    const req = 'https://data.grandlyon.com/api/elasticsearch/_search';
    const queryString = data.searchQuery.trim().replace(/\s/g, ' AND ');
    const params = {
      from: 0,
      size: 30,
      _source: ['data-fr'],
      query: {
        bool: {
          filter: {
            term: {
              'metadata-fr.geonet:info.uuid.keyword': '4cb035de-6ac3-4763-94d8-4c19b1d19607',
            },
          },
          must: [
            {
              query_string: {
                query: queryString,
                default_field: '*',
                analyzer: 'my_search_analyzer',
                fuzziness: 'AUTO',
                minimum_should_match: '90%',
              },
            },
          ],
        },
      },
    };
    return new Promise((resolve, reject) => {
      this.httpService
        .request({
          url: req,
          method: 'GET',
          headers: { 'Content-Type': 'application/json' },
          data: params,
        })
        .subscribe(
          (reply) => {
            return resolve(reply.data);
          },
          (err) => {
            Logger.error(`Request error: ${err.config.url}`, 'StructureService - search');
            Logger.error(err);
          }
        );
    });
  }

  /**
   * Count every value occurence of a given key
   * @param key structure key
   * @return [{id: 'key', count: 'value'}]
   */
  public async countByStructureKey(key: string): Promise<any> {
    const uniqueElements = await this.structureModel.distinct(key).exec();
    return await Promise.all(
      uniqueElements.map(async (value) => {
        return {
          id: value,
          count: await this.structureModel
            .countDocuments({ $and: [{ [key]: { $elemMatch: { $eq: value } }, deletedAt: { $exists: false } }] })
            .exec(),
        };
      })
    );
  }

  public getCoord(numero: string, address: string, zipcode: string): Observable<AxiosResponse<any>> {
    const req =
      'https://download.data.grandlyon.com/geocoding/photon/api' + '?q=' + numero + ' ' + address + ' ' + zipcode;
    Logger.log(`Request : ${req}`, 'StructureService - getCoord');
    return this.httpService.get(encodeURI(req));
  }

  public async deleteOne(id: string): Promise<Structure> {
    const structure = await this.structureModel.findById(Types.ObjectId(id)).exec();
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    structure.deletedAt = DateTime.local().setZone('Europe/Paris').toString();
    this.anonymizeStructure(structure).save();
    return structure;
  }

  private anonymizeStructure(structure: StructureDocument): StructureDocument {
    structure.contactName = '';
    structure.contactSurname = '';
    structure.contactPhone = '';
    structure.contactMail = '';
    structure.facebook = '';
    structure.twitter = '';
    structure.instagram = '';
    structure.website = '';
    structure.fonction = '';
    structure.gender = '';
    return structure;
  }
  
  @Cron(CronExpression.EVERY_DAY_AT_4AM)
  public async checkOutdatedStructuresInfo(): Promise<void> {
    const OUTDATED_MONT_TO_CHECK = 6;
    const structureList = await this.findAll();
    const local = DateTime.local().setZone('Europe/Paris');
    // Get outdated structures
    const filteredList = structureList.filter((structure) => {
      const updateDate = DateTime.fromFormat(structure.updatedAt.split(' GMT')[0], 'EEE MMM dd yyyy HH:mm:ss');
      const diff = local.diff(updateDate, ['months']);
      if (diff.values.months > OUTDATED_MONT_TO_CHECK) {
        return true;
      }
    });
    // Get owners of outdated structures
    const ownerList = await Promise.all(
      filteredList.map(async (structure) => {
        const owner = await this.userService.isStructureClaimed(structure._id.toString());
        if (owner) {
          return { structure: structure, owner: owner };
        }
      })
    );
    // Send email if possible and update user
    ownerList
      .filter((x) => x != undefined)
      .forEach(async (data) => {
        this.userService.findOne(data.owner.email).then((user) => {
          // If mail is already sent, do not resend
          if (user.structureOutdatedMailSent.includes(data.structure._id)) {
            return;
          } else {
            this.sendOutdatedEmailToUser(data.owner.email, data.structure.structureName);
            user.structureOutdatedMailSent.push(data.structure._id);
            user.save();
          }
        });
      });
  }

  /**
   * Generate activation token and send it to user by email, in order to validate
   * a new account.
   * @param user User
   */
  private async sendOutdatedEmailToUser(userEmail: string, structureName: string): Promise<any> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureOutdatedInfo.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureOutdatedInfo.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      name: structureName,
    });
    this.mailerService.send(userEmail, jsonConfig.subject, html);
  }
}
